from __future__ import annotations

from abc import ABC, abstractmethod


class VersionSchemeInterface(ABC):  # no cov
    """
    Example usage:

    === ":octicons-file-code-16: plugin.py"

        ```python
        from hatchling.version.scheme.plugin.interface import VersionSchemeInterface


        class SpecialVersionScheme(VersionSchemeInterface):
            PLUGIN_NAME = 'special'
            ...
        ```

    === ":octicons-file-code-16: hooks.py"

        ```python
        from hatchling.plugin import hookimpl

        from .plugin import SpecialVersionScheme


        @hookimpl
        def hatch_register_version_scheme():
            return SpecialVersionScheme
        ```
    """

    PLUGIN_NAME = ''
    """The name used for selection."""

    def __init__(self, root, config):
        self.__root = root
        self.__config = config

    @property
    def root(self):
        """
        The root of the project tree as a string.
        """
        return self.__root

    @property
    def config(self):
        """
        === ":octicons-file-code-16: pyproject.toml"

            ```toml
            [tool.hatch.version]
            ```

        === ":octicons-file-code-16: hatch.toml"

            ```toml
            [version]
            ```
        """
        return self.__config

    @abstractmethod
    def update(self, desired_version: str, original_version: str, version_data: dict):
        """
        This should return a normalized form of the desired version and verify that it
        is higher than the original version.
        """



import os
import pickle
import sys


class InvokedApplication:
    def display_info(self, *args, **kwargs):
        send_app_command('display_info', *args, **kwargs)

    def display_waiting(self, *args, **kwargs):
        send_app_command('display_waiting', *args, **kwargs)

    def display_success(self, *args, **kwargs):
        send_app_command('display_success', *args, **kwargs)

    def display_warning(self, *args, **kwargs):
        send_app_command('display_warning', *args, **kwargs)

    def display_error(self, *args, **kwargs):
        send_app_command('display_error', *args, **kwargs)

    def display_debug(self, *args, **kwargs):
        send_app_command('display_debug', *args, **kwargs)

    def display_mini_header(self, *args, **kwargs):
        send_app_command('display_mini_header', *args, **kwargs)

    def abort(self, *args, **kwargs):
        send_app_command('abort', *args, **kwargs)
        sys.exit(kwargs.get('code', 1))

    def get_safe_application(self):
        return SafeApplication(self)


class Application:
    """
    The way output is displayed can be [configured](../config/hatch.md#terminal) by users.

    !!! important
        Never import this directly; Hatch judiciously decides if a type of plugin requires
        the capabilities herein and will grant access via an attribute.
    """

    def __init__(self):
        self.__verbosity = int(os.environ.get('HATCH_VERBOSE', '0')) - int(os.environ.get('HATCH_QUIET', '0'))

    def display_info(self, message='', **kwargs):
        """
        Meant to be used for messages conveying basic information.
        """
        if self.__verbosity >= 0:
            print(message)

    def display_waiting(self, message='', **kwargs):
        """
        Meant to be used for messages shown before potentially time consuming operations.
        """
        if self.__verbosity >= 0:
            print(message)

    def display_success(self, message='', **kwargs):
        """
        Meant to be used for messages indicating some positive outcome.
        """
        if self.__verbosity >= 0:
            print(message)

    def display_warning(self, message='', **kwargs):
        """
        Meant to be used for messages conveying important information.
        """
        if self.__verbosity >= -1:
            print(message)

    def display_error(self, message='', **kwargs):
        """
        Meant to be used for messages indicating some unrecoverable error.
        """
        if self.__verbosity >= -2:
            print(message)

    def display_debug(self, message='', level=1, **kwargs):
        """
        Meant to be used for messages that are not useful for most user experiences.
        The `level` option must be between 1 and 3 (inclusive).
        """
        if not 1 <= level <= 3:
            raise ValueError('Debug output can only have verbosity levels between 1 and 3 (inclusive)')
        elif self.__verbosity >= level:
            print(message)

    def display_mini_header(self, message='', **kwargs):
        if self.__verbosity >= 0:
            print(f'[{message}]')

    def abort(self, message='', code=1, **kwargs):
        """
        Terminate the program with the given return code.
        """
        if message and self.__verbosity >= -2:
            print(message)

        sys.exit(code)

    def get_safe_application(self):
        return SafeApplication(self)


class SafeApplication:
    def __init__(self, app):
        self.abort = app.abort
        self.display_info = app.display_info
        self.display_error = app.display_error
        self.display_success = app.display_success
        self.display_waiting = app.display_waiting
        self.display_warning = app.display_warning
        self.display_debug = app.display_debug
        self.display_mini_header = app.display_mini_header


def format_app_command(method, *args, **kwargs):
    procedure = pickle.dumps((method, args, kwargs), 4)

    return f"__HATCH__:{''.join('%02x' % i for i in procedure)}"


def get_application(called_by_app):
    return InvokedApplication() if called_by_app else Application()


def send_app_command(method, *args, **kwargs):
    _send_app_command(format_app_command(method, *args, **kwargs))


def _send_app_command(command):
    print(command)



import os
import string
from abc import ABC, abstractmethod
from collections import ChainMap
from contextlib import contextmanager

from hatchling.utils.fs import path_to_uri


class ContextFormatter(ABC):
    @abstractmethod
    def get_formatters(self):
        """
        This returns a mapping of supported field names to their respective formatting functions. Each function
        accepts 2 arguments:

        - the `value` that was passed to the format call, defaulting to `None`
        - the modifier `data`, defaulting to an empty string
        """

    @classmethod
    def format_path(cls, path, modifier):
        if not modifier:
            return os.path.normpath(path)
        elif modifier == 'uri':
            return path_to_uri(path)
        elif modifier == 'real':
            return os.path.realpath(path)
        else:
            raise ValueError(f'Unknown path modifier: {modifier}')


class DefaultContextFormatter(ContextFormatter):
    CONTEXT_NAME = 'default'

    def __init__(self, root):
        self.__root = root

    def get_formatters(self):
        return {
            '/': self.__format_directory_separator,
            ';': self.__format_path_separator,
            'env': self.__format_env,
            'home': self.__format_home,
            'root': self.__format_root,
        }

    def __format_directory_separator(self, value, data):
        return os.sep

    def __format_path_separator(self, value, data):
        return os.pathsep

    def __format_root(self, value, data):
        return self.format_path(self.__root, data)

    def __format_home(self, value, data):
        return self.format_path(os.path.expanduser('~'), data)

    def __format_env(self, value, data):
        if not data:
            raise ValueError('The `env` context formatting field requires a modifier')

        env_var, separator, default = data.partition(':')
        if env_var in os.environ:
            return os.environ[env_var]
        elif not separator:
            raise ValueError(f'Nonexistent environment variable must set a default: {env_var}')
        else:
            return default


class Context:
    def __init__(self, root):
        self.__root = str(root)

        # Allow callers to define their own formatters with precedence
        self.__formatters = ChainMap()
        self.__configured_contexts = set()
        self.__formatter = ContextStringFormatter(self.__formatters)

        self.add_context(DefaultContextFormatter(self.__root))

    def format(self, *args, **kwargs):
        return self.__formatter.format(*args, **kwargs)

    def add_context(self, context):
        if context.CONTEXT_NAME in self.__configured_contexts:
            return

        self.__add_formatters(context.get_formatters())
        self.__configured_contexts.add(context.CONTEXT_NAME)

    @contextmanager
    def apply_context(self, context):
        self.__add_formatters(context.get_formatters())
        try:
            yield
        finally:
            self.__remove_formatters()

    def __add_formatters(self, formatters):
        return self.__formatters.maps.insert(0, formatters)

    def __remove_formatters(self):
        if len(self.__formatters.maps) > 1:
            self.__formatters.maps.pop(0)


class ContextStringFormatter(string.Formatter):
    def __init__(self, formatters):
        super().__init__()

        self.__formatters = formatters

    def vformat(self, format_string, args, kwargs):
        # We override to increase the recursion limit from 2 to 10
        used_args = set()
        result, _ = self._vformat(format_string, args, kwargs, used_args, 10)
        self.check_unused_args(used_args, args, kwargs)
        return result

    def get_value(self, key, args, kwargs):
        if key in self.__formatters:
            # Avoid hard look-up and rely on `None` to indicate that the field is undefined
            return kwargs.get(key)
        else:
            try:
                return super().get_value(key, args, kwargs)
            except KeyError:
                raise ValueError(f'Unknown context field `{key}`') from None

    def format_field(self, value, format_spec):
        formatter, _, data = format_spec.partition(':')
        if formatter in self.__formatters:
            return self.__formatters[formatter](value, data)
        else:
            return super().format_field(value, format_spec)

    def parse(self, format_string):
        for literal_text, field_name, format_spec, conversion in super().parse(format_string):
            if field_name in self.__formatters:
                yield literal_text, field_name, f'{field_name}:{format_spec}', conversion
            else:
                yield literal_text, field_name, format_spec, conversion



from hatchling.plugin import hookimpl
from hatchling.version.source.code import CodeSource
from hatchling.version.source.regex import RegexSource


@hookimpl
def hatch_register_version_source():
    return [CodeSource, RegexSource]



def load_plugin_from_script(path, script_name, plugin_class, plugin_id):
    import importlib

    spec = importlib.util.spec_from_file_location(script_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    plugin_finder = f'get_{plugin_id}'
    names = dir(module)
    if plugin_finder in names:
        return getattr(module, plugin_finder)()

    subclasses = []
    for name in names:
        obj = getattr(module, name)
        if obj is plugin_class:
            continue

        try:
            if issubclass(obj, plugin_class):
                subclasses.append(obj)
        except TypeError:
            continue

    if not subclasses:
        raise ValueError(f'Unable to find a subclass of `{plugin_class.__name__}` in `{script_name}`: {path}')
    elif len(subclasses) > 1:
        raise ValueError(
            f'Multiple subclasses of `{plugin_class.__name__}` found in `{script_name}`, '
            f'select one by defining a function named `{plugin_finder}`: {path}'
        )
    else:
        return subclasses[0]



class BuildHookInterface:  # no cov
    """
    Example usage:

    === ":octicons-file-code-16: plugin.py"

        ```python
        from hatchling.builders.hooks.plugin.interface import BuildHookInterface


        class SpecialBuildHook(BuildHookInterface):
            PLUGIN_NAME = 'special'
            ...
        ```

    === ":octicons-file-code-16: hooks.py"

        ```python
        from hatchling.plugin import hookimpl

        from .plugin import SpecialBuildHook


        @hookimpl
        def hatch_register_build_hook():
            return SpecialBuildHook
        ```
    """

    PLUGIN_NAME = ''
    """The name used for selection."""

    def __init__(self, root, config, build_config, metadata, directory, target_name, app=None):
        self.__root = root
        self.__config = config
        self.__build_config = build_config
        self.__metadata = metadata
        self.__directory = directory
        self.__target_name = target_name
        self.__app = app

    @property
    def app(self):
        """
        An instance of [Application](../utilities.md#hatchling.bridge.app.Application).
        """
        if self.__app is None:
            from hatchling.bridge.app import Application

            self.__app = Application().get_safe_application()

        return self.__app

    @property
    def root(self):
        """
        The root of the project tree.
        """
        return self.__root

    @property
    def config(self):
        """
        The cumulative hook configuration.

        === ":octicons-file-code-16: pyproject.toml"

            ```toml
            [tool.hatch.build.hooks.<PLUGIN_NAME>]
            [tool.hatch.build.targets.<TARGET_NAME>.hooks.<PLUGIN_NAME>]
            ```

        === ":octicons-file-code-16: hatch.toml"

            ```toml
            [build.hooks.<PLUGIN_NAME>]
            [build.targets.<TARGET_NAME>.hooks.<PLUGIN_NAME>]
            ```
        """
        return self.__config

    @property
    def metadata(self):
        # Undocumented for now
        return self.__metadata

    @property
    def build_config(self):
        """
        An instance of [BuilderConfig](../utilities.md#hatchling.builders.config.BuilderConfig).
        """
        return self.__build_config

    @property
    def directory(self):
        """
        The build directory.
        """
        return self.__directory

    @property
    def target_name(self):
        """
        The plugin name of the build target.
        """
        return self.__target_name

    def clean(self, versions):
        """
        This occurs before the build process if the `-c`/`--clean` flag was passed to
        the [`build`](../../cli/reference.md#hatch-build) command, or when invoking
        the [`clean`](../../cli/reference.md#hatch-clean) command.
        """

    def initialize(self, version, build_data):
        """
        This occurs immediately before each build.

        Any modifications to the build data will be seen by the build target.
        """

    def finalize(self, version, build_data, artifact_path):
        """
        This occurs immediately after each build and will not run if the `--hooks-only` flag
        was passed to the [`build`](../../cli/reference.md#hatch-build) command.

        The build data will reflect any modifications done by the target during the build.
        """
